using System;
using System.Windows.Forms;
using System.Drawing;

namespace DockApp
{
	/// <summary>
	/// Base class for dockapps.
	/// </summary>
	public class DockApp : Form
	{
		/// <summary>
		/// A canvas object on which the dockapp should draw
		/// </summary>
		protected Panel canvas;
		/// <summary>
		/// Context menu for the application - default has About and Close.
		/// </summary>
		protected ContextMenu contextMenu;
		/// <summary>
		/// A Bitmap for drawing things off screen. Used in double buffering.
		/// </summary>
		protected Bitmap offScreen;
		/// <summary>
		/// Graphics object to draw with for offscreen graphics. Used in double buffering.
		/// </summary>
		protected Graphics offScreenGraphics;
		/// <summary>
		/// Copy of the Graphics context for this window. If you are using double buffering,
		/// do not draw directly to this object.
		/// </summary>
		protected Graphics graphics;
		private Rectangle screenSize;
		private MenuItem toggleOnTopMenuItem;
		
		public DockApp()
		{
			InitializeForm();

			offScreen = new Bitmap(canvas.Width, canvas.Height);
			offScreenGraphics = Graphics.FromImage(offScreen);

			graphics = Graphics.FromHwnd(canvas.Handle);
		}

		private String aboutText = "Generic Windows DockApp\nFramework by Nelson Minar <nelson@monkey.org>";
		protected String AboutText 
		{
			get 
			{
				return aboutText;
			}
			set 
			{
				aboutText = value;
			}
		}

		/// <summary>
		/// Initialize the DockApp Form to the standard DockApp
		/// </summary>
		virtual protected void InitializeForm() 
		{
			int width = 75;			// 75 pixels is the default Windows desktop grid size
			int height = 75;

			this.SuspendLayout();

			// Figure out the screen size
			Screen s = Screen.PrimaryScreen;
			screenSize = s.Bounds;

			// Configure the top level form
			// The size of a DockApp is always fixed
			this.AutoScale = false;
			this.ClientSize = new System.Drawing.Size(width, height);
			this.MinimumSize = new System.Drawing.Size(1, 1);

			// DockApp has no Windows border or frame
			this.FormBorderStyle = FormBorderStyle.None;

			// No taskbar icon, either.
			this.ShowInTaskbar = false;

			// Start the window on top
			this.TopMost = true;

			// Decorate the form
			this.Name = "DockApp";
			this.Text = "Generic DockApp";

			// Construct a context menu and assign it to the top level form
			contextMenu = new ContextMenu();
			contextMenu.MenuItems.Add("About", new EventHandler(aboutItem_Click));
			toggleOnTopMenuItem = new MenuItem("Always on top", new EventHandler(toggleOnTopItem_Click));
			toggleOnTopMenuItem.Checked = this.TopMost;
			contextMenu.MenuItems.Add(toggleOnTopMenuItem);
			contextMenu.MenuItems.Add("Close", new EventHandler(closeItem_Click));
			this.ContextMenu = contextMenu;

			// Configure the canvas for drawing
			canvas = new Panel();
			canvas.Name = "canvas";

			// Decorations for the canvas
			canvas.BackColor = Color.Black;
			canvas.BorderStyle = BorderStyle.Fixed3D;

			// Set its location when it goes in the parent
			canvas.Location = new Point(1, 1);
			canvas.Size = new Size(width-2, height-2);

			// Add the canvas to the top level form
			this.Controls.Add(canvas);

			InstallEventHandlers();

			this.ResumeLayout();
		}

		/// <summary>
		/// Install default event handlers for a DockApp. Handlers are installed on the canvas.
		/// Handlers for Paint, MouseDown, MouseMove, and MouseUp.
		/// </summary>
		virtual protected void InstallEventHandlers() 
		{
			canvas.Paint += new PaintEventHandler(paint);
			canvas.MouseDown += new MouseEventHandler(beginMoveWindow);
			canvas.MouseMove += new MouseEventHandler(moveWindow);
			canvas.MouseUp += new MouseEventHandler(endMoveWindow);
		}

		/// <summary>
		/// Draw the contents of the offscreen buffer onto the main screen.
		/// </summary>
		protected void paintOffScreen() 
		{
			graphics.DrawImage(offScreen, 0, 0);
		}

		/// <summary>
		/// Render the image on screen. This is installed as a Paint event ahnderl. By default this copies the offscreen image onto the
		/// canvas. If you override this, best to draw on the offscreen bitmap and then call the base method at the end.
		/// </summary>
		virtual protected void paint(object sender, PaintEventArgs e) 
		{
			paintOffScreen();
		}


		/// <summary>
		/// If true, then we're in the middle of moving the window.
		/// </summary>
		protected bool windowMoving = false;
		/// <summary>
		/// If we're moving the window, moveOrigin stores where the mouse was to begin with.
		/// </summary>
		protected Point moveOrigin;

		/// <summary>
		/// Handle MouseDown events to begin a window move. If the button is not the left button, forward it on up
		/// to the form. Otherwise begin a window move.
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		virtual protected void beginMoveWindow(Object sender, MouseEventArgs e) 
		{
			if (e.Button != MouseButtons.Left)
				this.OnMouseDown(e);
			else 
			{
				windowMoving = true;
				moveOrigin = new Point(e.X, e.Y);
			}
		}

		/// <summary>
		/// Handle MouseMove events for moving windows. If we're not in move mode, then forward event up to form.
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		virtual protected void moveWindow(Object sender, MouseEventArgs e) 
		{
			if (!windowMoving)
				this.OnMouseMove(e);
			else 
			{
				this.Location = calculateNewLocation(e);
			}
		}

		/// <summary>
		/// Handle MouseUp events for moving windows. If we're not in move mode, then forward event up to form.
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		virtual protected void endMoveWindow(Object sender, MouseEventArgs e) 
		{
			if (!windowMoving)
				this.OnMouseUp(e);
			else 
			{
				this.Location = calculateNewLocation(e);
				windowMoving = false;
			}
		}

		/// <summary>
		/// Snap the location of the window if it gets near the edge.
		/// </summary>
		/// <param name="e"></param>
		/// <returns></returns>
		private Point calculateNewLocation(MouseEventArgs e) 
		{
			int newX = this.Location.X + e.X - moveOrigin.X;
			int newY = this.Location.Y + e.Y - moveOrigin.Y;

			if (newX > screenSize.Width-(this.Width*3/2)) 
			{
				newX = screenSize.Width-this.Width;
				newY = (newY / this.Width) * this.Width;
			} 
			else if (newX < screenSize.X + (this.Width/2)) 
			{
				newX = 0;
				newY = (newY / this.Width) * this.Width;
			}
			return new Point(newX, newY);
		}

		/// <summary>
		/// A handler for when the "About" context menu item is chosen.
		/// Default behaviour is to show the contents of AboutText.
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		/// 

		virtual protected void aboutItem_Click(Object sender, EventArgs e) 
		{
			MessageBox.Show(AboutText);
		}

		/// <summary>
		/// A handler for when the "Exit" context menu item is chosen
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		virtual protected void closeItem_Click(Object sender, EventArgs e) 
		{
			Application.Exit();
		}

		virtual protected void toggleOnTopItem_Click(Object sender, EventArgs e) 
		{
			this.TopMost = !this.TopMost;
			toggleOnTopMenuItem.Checked = this.TopMost;
		}
	}
}
